/*********************************************
 * vim:sw=8:ts=8:si:et
 * To use the above modeline in vim you must have "set modeline" in your .vimrc
 * Author: Guido Socher
 * Copyright: GPL V2
 * See http://www.gnu.org/licenses/gpl.html
 *
 * Ethernet remote device and sensor
 *
 * Chip type           : ATMEGA88/ATMEGA168/atmega328 with ENC28J60
 *********************************************/
#include <avr/io.h>
#include <stdlib.h>
#include <string.h>
#include <avr/pgmspace.h>
#include <avr/interrupt.h>
#include "ip_arp_udp_tcp.h"
#include "enc28j60.h"
#include "timeout.h"
#include "avr_compat.h"
#include "net.h"
#include "sensirion_protocol.h"

#define VERSIONSTRING "<br><hr><br>version 3.2</small>\n"

// please modify the following two lines. mac and ip have to be unique
// in your local area network. You can not have the same numbers in
// two devices:
//static uint8_t mymac[6] = {0x54,0x55,0x58,0x10,0x01,0x27};
// how did I get the mac addr? Translate the first 3 numbers into ascii is: TUX
//static uint8_t myip[4] = {192,168,1,27};

static uint8_t mymac[6] = {0x54,0x55,0x58,0x10,0x00,0x26};
// how did I get the mac addr? Translate the first 3 numbers into ascii is: TUX
static uint8_t myip[4] = {10,0,0,26};
//
// listen port for tcp/www 
#define MYWWWPORT 80 
// or on a different port:
// listen port for tcp/www 
//#define MYWWWPORT 88  
//
// bigger pages need more buffer but we are limited be the amount of ram
#define BUFFER_SIZE 460
static uint8_t buf[BUFFER_SIZE+1];

// uncomment to see raw sensor data:
//#define SHOW_RAW 1
static uint8_t s_conreset_cnt=0;  // do not reset the sensirion connection everytime

// gerneral purpose LED
/* set output to Vcc, red LED off */
#define LED1OFF PORTB|=(1<<PORTB1)
/* set output to GND, red LED on */
#define LED1ON PORTB&=~(1<<PORTB1)
// 
// 

// convert temperature to kelvin times 10
int c2f(int celsiustimes10){
        return((celsiustimes10 * 18 + 3200)/10);
}

// insert a '.' before the last digit/character in the string.
// The user is responsible for allocating enough space.
// "100" will become "10.0"
// Special care must be taken if there is just one digit: -5-> -0.5
void adddecimalpoint(char *str){
        char c;
        if (*str == '-'){
                str++;
        }
        if (*(str+1) == '\0'){
                // just one digit
                c=*str;
                *str='0';
                *(str+1)='.';
                *(str+2)=c;
                *(str+3)='\0';
                return;
        }
        while(*str){
                c=*str;
                str++;
                if (!*str){
                        // last char, insert a dot
                        *(str-1)='.';
                        *str=c;
                        str++;
                        *str='\0';
                }
        }
}

// prepare the webpage by writing the data to the tcp send buffer
uint16_t print_webpage(uint8_t *buf)
{
        uint16_t plen;
        unsigned char error;
        unsigned char rh;
        unsigned int humival_raw,tempval_raw;
        int dew,temp; 
        char str[15];
        LED1ON;
        // autmatically refresh page every 5 min:
        plen=fill_tcp_data_p(buf,0,PSTR("HTTP/1.0 200 OK\r\nContent-Type: text/html\r\nPragma: no-cache\r\nRefresh: 300\r\n\r\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR("<h2>Sensor @"));
        itoa(myip[3],str,10);
        plen=fill_tcp_data(buf,plen,str);
        plen=fill_tcp_data_p(buf,plen,PSTR("</h2>\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR("<pre>\n"));
        if (s_conreset_cnt==0){
                s_connectionreset();
                s_conreset_cnt++;
        }
        if (s_conreset_cnt>9){
                s_conreset_cnt=0;
        }
        error=s_measure( &tempval_raw,0); //measure temperature
        if (error==0){
                error=s_measure( &humival_raw,1); //measure humidity
        }
        //
        if(error!=0){ //in case of an error
                if (error==1){
                        plen=fill_tcp_data_p(buf,plen,PSTR("sensor communication error\n"));
                }
                if (error==2){
                        plen=fill_tcp_data_p(buf,plen,PSTR("sensor crc error\n"));
                }
                if (error==3){
                        plen=fill_tcp_data_p(buf,plen,PSTR("sensor timeout error\n"));
                }
                itoa(humival_raw,str,10);
                plen=fill_tcp_data(buf,plen,str);
                plen=fill_tcp_data_p(buf,plen,PSTR("\n\n"));
                plen=fill_tcp_data_p(buf,plen,PSTR("</pre>\n"));
                s_conreset_cnt=0;
                return(plen);
        }
#ifdef SHOW_RAW
        plen=fill_tcp_data_p(buf,plen,PSTR("DEBUG: raw temp humi:"));
        itoa(tempval_raw,str,10);
        plen=fill_tcp_data(buf,plen,str);
        plen=fill_tcp_data_p(buf,plen,PSTR(" "));
        itoa(humival_raw,str,10);
        plen=fill_tcp_data(buf,plen,str);
        plen=fill_tcp_data_p(buf,plen,PSTR("\n\n"));
#endif
        //
        plen=fill_tcp_data_p(buf,plen,PSTR("temperature  : "));
        temp=calc_sth11_temp(tempval_raw);
        itoa(temp,str,10);
        adddecimalpoint(str);
        plen=fill_tcp_data(buf,plen,str);
        plen=fill_tcp_data_p(buf,plen,PSTR("'C ["));
        itoa(c2f(temp),str,10);
        adddecimalpoint(str);
        plen=fill_tcp_data(buf,plen,str);
        plen=fill_tcp_data_p(buf,plen,PSTR("'F]"));
#ifdef SHOW_RAW
        plen=fill_tcp_data_p(buf,plen,PSTR("\nhumi lin     : "));
        rh=rhcalc_int(humival_raw);
        itoa(rh,str,10);
        plen=fill_tcp_data(buf,plen,str);
#endif
        plen=fill_tcp_data_p(buf,plen,PSTR("\nrel. humidity: "));
        rh=calc_sth11_humi(humival_raw,temp);
        itoa(rh,str,10);
        plen=fill_tcp_data(buf,plen,str);
        plen=fill_tcp_data_p(buf,plen,PSTR("%\ndew point    : "));
        dew=calc_dewpoint(rh,temp);
        itoa(dew,str,10);
        adddecimalpoint(str);
        plen=fill_tcp_data(buf,plen,str);
        plen=fill_tcp_data_p(buf,plen,PSTR("'C ["));
        // cut comma as the comma digits are not accurate:
        itoa(c2f(dew)/10,str,10);
        plen=fill_tcp_data(buf,plen,str);
        plen=fill_tcp_data_p(buf,plen,PSTR("'F]\n"));
        //
        plen=fill_tcp_data_p(buf,plen,PSTR("</pre>\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR("<small><br>"));
        plen=fill_tcp_data_p(buf,plen,PSTR("<a href=\"./\">refresh page</a>\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR(VERSIONSTRING));
        return(plen);
}


int main(void){

        
        uint16_t plen;
        uint16_t dat_p;
        uint8_t i=0;
        
        // set the clock speed to "no pre-scaler" (8MHz with internal osc or 
        // full external speed)
        // set the clock prescaler. First write CLKPCE to enable setting of clock the
        // next four instructions.
        CLKPR=(1<<CLKPCE); // change enable
        CLKPR=0; // "no pre-scaler"
        _delay_ms(1);

        /* enable PD2/INT0, as input */
        DDRD&= ~(1<<DDD2);

        /*initialize enc28j60*/
        enc28j60Init(mymac);
        enc28j60clkout(2); // change clkout from 6.25MHz to 12.5MHz
        _delay_ms(10);
        
        // LED
        /* enable PB1, LED as output */
        DDRB|= (1<<DDB1);

        LED1OFF;

        // the transistor on PD7
        DDRD|= (1<<DDD7);
        PORTD &= ~(1<<PORTD7);// transistor off
        
        /* Magjack leds configuration, see enc28j60 datasheet, page 11 */
        // LEDB=yellow LEDA=green
        //
        // 0x476 is PHLCON LEDA=links status, LEDB=receive/transmit
        // enc28j60PhyWrite(PHLCON,0b0000 0100 0111 01 10);
        enc28j60PhyWrite(PHLCON,0x476);
        _delay_ms(20);
        
        i=1;

        //init the ethernet/ip layer:
        init_ip_arp_udp_tcp(mymac,myip,MYWWWPORT);

        while(1){
                // get the next new packet:
                plen = enc28j60PacketReceive(BUFFER_SIZE, buf);

                /*plen will ne unequal to zero if there is a valid 
                 * packet (without crc error) */
                if(plen==0){
                        continue;
                }
                        
                // arp is broadcast if unknown but a host may also
                // verify the mac address by sending it to 
                // a unicast address.
                if(eth_type_is_arp_and_my_ip(buf,plen)){
                        make_arp_answer_from_request(buf);
                        continue;
                }

                // check if ip packets are for us:
                if(eth_type_is_ip_and_my_ip(buf,plen)==0){
                        continue;
                }
                
                if(buf[IP_PROTO_P]==IP_PROTO_ICMP_V && buf[ICMP_TYPE_P]==ICMP_TYPE_ECHOREQUEST_V){
                        // a ping packet, let's send pong
                        make_echo_reply_from_request(buf,plen);
                        continue;
                }
                // tcp port www start, compare only the lower byte
                if (buf[IP_PROTO_P]==IP_PROTO_TCP_V&&buf[TCP_DST_PORT_H_P]==0&&buf[TCP_DST_PORT_L_P]==MYWWWPORT){
                        if (buf[TCP_FLAGS_P] & TCP_FLAGS_SYN_V){
                                make_tcp_synack_from_syn(buf);
                                // make_tcp_synack_from_syn does already send the syn,ack
                                continue;
                        }
                        if (buf[TCP_FLAGS_P] & TCP_FLAGS_ACK_V){
                                init_len_info(buf); // init some data structures
                                // we can possibly have no data, just ack:
                                dat_p=get_tcp_data_pointer();
                                if (dat_p==0){
                                        if (buf[TCP_FLAGS_P] & TCP_FLAGS_FIN_V){
                                                // finack, answer with ack
                                                make_tcp_ack_from_any(buf);
                                        }
                                        // just an ack with no data, wait for next packet
                                        continue;
                                }
                                if (strncmp("GET ",(char *)&(buf[dat_p]),4)!=0){
                                        // head, post and other methods:
                                        //
                                        // for possible status codes see:
                                        // http://www.w3.org/Protocols/rfc2616/rfc2616-sec10.html
                                        plen=fill_tcp_data_p(buf,0,PSTR("HTTP/1.0 200 OK\r\nContent-Type: text/html\r\n\r\n<h1>200 OK</h1>"));
                                        goto SENDTCP;
                                }
                                if (strncmp("/ ",(char *)&(buf[dat_p+4]),2)==0){
                                        plen=print_webpage(buf);
                                        goto SENDTCP;
                                }
                                // not the correct url. In this case we 
                                // just accept the base url (hostname/ip only).
                                // Generate and error.
                                //
                                // for possible status codes see:
                                // http://www.w3.org/Protocols/rfc2616/rfc2616-sec10.html
                                // 401 Unauthorized
                                // 403 Forbidden
                                // 404 Not Found
                                // 410 Gone
                                //
                                //plen=fill_tcp_data_p(buf,0,PSTR("HTTP/1.0 401 Unauthorized\r\nContent-Type: text/html\r\n\r\n<h1>401 Unauthorized</h1>"));
                                plen=fill_tcp_data_p(buf,0,PSTR("HTTP/1.0 404 Not Found\r\nContent-Type: text/html\r\n\r\n<h1>404 Not Found</h1>"));
                                //
SENDTCP:
                                make_tcp_ack_from_any(buf); // send ack for http get
                                make_tcp_ack_with_data(buf,plen); // send data
                                LED1OFF;
                                continue;
                        }

                }
                // tcp port www end
                //
        }
        return (0);
}
